import psutil
import os
import threading
import gc
from sys import getswitchinterval

from opentelemetry import metrics
from opentelemetry.metrics import CallbackOptions, Observation
# from opentelemetry.sdk.resources import Resource
# from opentelemetry.sdk.metrics import MeterProvider
# from opentelemetry.sdk.metrics.export import PeriodicExportingMetricReader
# from opentelemetry.exporter.otlp.proto.grpc.metric_exporter import OTLPMetricExporter

mw_agent_target = os.environ.get('MW_AGENT_SERVICE', '127.0.0.1')


def collect_metrics() -> None:
    # metrics.set_meter_provider(MeterProvider(
    #     resource=Resource.create({
    #         "service.name": service_name,
    #         "project.name": project_name,
    #         "mw.app.lang": "python",
    #         "runtime.metrics.python": "true"
    #     }),
    #     metric_readers=[
    #         PeriodicExportingMetricReader(
    #             OTLPMetricExporter(insecure=True, endpoint=mw_agent_target + ":9319")
    #         )
    #     ],
    # ))
    provider = metrics.get_meter_provider()
    meter = provider.get_meter("middleware-apm", "0.2.0")
    _generate_metrics(meter)


def _generate_metrics(meter):
    meter.create_observable_gauge(
        "process.cpu_usage.percentage",
        unit="Percent",
        callbacks=[_cpu_usage_cb],
        description="The will show the cpu usage of the process"
    )
    meter.create_observable_gauge(
        "process.memory_usage.percentage",
        unit="Percent",
        callbacks=[_ram_usage_cb],
        description="The will show the memory usage of the process"
    )
    meter.create_observable_counter(
        "process.cpu.time",
        callbacks=[_cpu_time_callback],
        unit="s",
        description="This will show the cpu time of the process"
    )
    meter.create_observable_gauge(
        "process.memory.bytes",
        unit="Bytes",
        callbacks=[_memory_rss_cb, _memory_vms_cb, _memory_shared_cb, _memory_text_cb, _memory_data_cb],
        description="The will show the memory rss, vms, shared, text and data of the process"
    )
    meter.create_observable_gauge(
        "process.num_threads.count",
        unit="Count",
        callbacks=[_num_threads_cb],
        description="The will show the number of threads of the process"
    )
    meter.create_observable_gauge(
        "process.num_fds.count",
        unit="Count",
        callbacks=[_num_fds_cb],
        description="The will show the number of fds of the process"
    )
    meter.create_observable_gauge(
        "process.ctx_switches.count",
        unit="Count",
        callbacks=[_num_vctx_switches_cb, _num_ivctx_switches_cb],
        description="The will show the number of ctx switches of the process"
    )
    meter.create_observable_gauge(
        "process.num_connections.count",
        unit="Count",
        callbacks=[_num_connections_cb],
        description="The will show the number of connections of the process"
    )
    meter.create_observable_gauge(
        "process.io_counters.bytes",
        unit="Bytes",
        callbacks=[_io_read_bytes_cb, _io_write_bytes_cb],
        description="This will show the io read and write bytes of the process"
    )
    meter.create_observable_gauge(
        "process.cpu_affinity.count",
        unit="Count",
        callbacks=[_cpu_affinity_cb],
        description="The will show the cpu affinity of the process"
    )
    meter.create_observable_gauge(
        "process.create_time.timestamp",
        unit="s",
        callbacks=[_create_time_cb],
        description="The will show the create time of the process"
    )
    meter.create_observable_gauge(
        "process.open_files.count",
        unit="Count",
        callbacks=[_open_file_counts_cb],
        description="The will show the open files of the process"
    )
    meter.create_observable_gauge(
        "process.threads.count",
        unit="Count",
        callbacks=[_thread_counts_cb],
        description="The will show the threads of the process"
    )
    meter.create_observable_gauge(
        "process.gc.count",
        unit="Count",
        callbacks=[_gc0_cb, _gc1_cb, _gc2_cb],
        description="The will show the gc of the process"
    )
    meter.create_observable_gauge(
        "process.context_switches.count",
        unit="Count",
        callbacks=[_context_switch_cb],
        description="The will show the context switches of the process"
    )
    meter.create_observable_gauge(
        "process.disk_usage.percentage",
        unit="Percent",
        callbacks=[_disk_usage_percent_cb],
        description="The will show the disk usage percentage of the process"
    )
    meter.create_observable_gauge(
        "process.disk_usage.bytes",
        unit="Bytes",
        callbacks=[_disk_usage_total_cb, _disk_usage_used_cb, _disk_usage_free_cb],
        description="The will show the disk usage of the process"
    )


def _cpu_usage_cb(options: CallbackOptions):
    print("--- Metrics Generated ---")
    yield Observation(value=psutil.Process(os.getpid()).cpu_percent())


def _ram_usage_cb(options: CallbackOptions):
    yield Observation(value=psutil.Process(os.getpid()).memory_percent())


def _cpu_time_callback(options: CallbackOptions):
    cpu_times = psutil.Process(os.getpid()).cpu_times()
    yield Observation(value=cpu_times.user, attributes={"state": "user"})
    yield Observation(value=cpu_times.system, attributes={"state": "system"})
    yield Observation(value=cpu_times.children_user, attributes={"state": "children_user"})
    yield Observation(value=cpu_times.children_system, attributes={"state": "children_system"})
    yield Observation(value=cpu_times.iowait, attributes={"state": "iowait"})


def _memory_rss_cb(options: CallbackOptions):
    yield Observation(
        value=psutil.Process(os.getpid()).memory_info().rss,
        attributes={"type": "rss"}
    )


def _memory_vms_cb(options: CallbackOptions):
    yield Observation(
        value=psutil.Process(os.getpid()).memory_info().vms,
        attributes={"type": "vms"}
    )


def _memory_shared_cb(options: CallbackOptions):
    yield Observation(
        value=psutil.Process(os.getpid()).memory_info().shared,
        attributes={"type": "shared"}
    )


def _memory_text_cb(options: CallbackOptions):
    yield Observation(
        value=psutil.Process(os.getpid()).memory_info().text,
        attributes={"type": "text"}
    )


def _memory_data_cb(options: CallbackOptions):
    yield Observation(
        value=psutil.Process(os.getpid()).memory_info().data,
        attributes={"type": "data"}
    )


def _num_threads_cb(options: CallbackOptions):
    yield Observation(value=psutil.Process(os.getpid()).num_threads())


def _num_fds_cb(options: CallbackOptions):
    yield Observation(value=psutil.Process(os.getpid()).num_fds())


def _num_vctx_switches_cb(options: CallbackOptions):
    yield Observation(
        value=psutil.Process(os.getpid()).num_ctx_switches().voluntary,
        attributes={"type": "voluntary"}
    )


def _num_ivctx_switches_cb(options: CallbackOptions):
    yield Observation(
        value=psutil.Process(os.getpid()).num_ctx_switches().involuntary,
        attributes={"type": "involuntary"}
    )


def _num_connections_cb(options: CallbackOptions):
    yield Observation(value=len(psutil.Process(os.getpid()).connections()))


def _io_read_bytes_cb(options: CallbackOptions):
    yield Observation(
        value=psutil.Process(os.getpid()).io_counters().read_bytes,
        attributes={"direction": "read"}
    )


def _io_write_bytes_cb(options: CallbackOptions):
    yield Observation(
        value=psutil.Process(os.getpid()).io_counters().write_bytes,
        attributes={"direction": "write"}
    )


def _cpu_affinity_cb(options: CallbackOptions):
    yield Observation(value=len(psutil.Process(os.getpid()).cpu_affinity()))


def _create_time_cb(options: CallbackOptions):
    yield Observation(value=psutil.Process(os.getpid()).create_time())


def _open_file_counts_cb(options: CallbackOptions):
    yield Observation(value=len(psutil.Process(os.getpid()).open_files()))


def _thread_counts_cb(options: CallbackOptions):
    yield Observation(value=threading.active_count())


def _gc0_cb(options: CallbackOptions):
    yield Observation(value=gc.get_count()[0], attributes={"type": "gc0"})


def _gc1_cb(options: CallbackOptions):
    yield Observation(value=gc.get_count()[1], attributes={"type": "gc1"})


def _gc2_cb(options: CallbackOptions):
    yield Observation(value=gc.get_count()[2], attributes={"type": "gc2"})


def _context_switch_cb(options: CallbackOptions):
    yield Observation(value=getswitchinterval())


def _disk_usage_percent_cb(options: CallbackOptions):
    yield Observation(value=psutil.disk_usage(os.sep).percent)


def _disk_usage_total_cb(options: CallbackOptions):
    yield Observation(value=psutil.disk_usage(os.sep).total, attributes={"type": "total"})


def _disk_usage_used_cb(options: CallbackOptions):
    yield Observation(value=psutil.disk_usage(os.sep).used, attributes={"type": "used"})


def _disk_usage_free_cb(options: CallbackOptions):
    yield Observation(value=psutil.disk_usage(os.sep).free, attributes={"type": "free"})



# import os
from opentelemetry import trace

# mw_agent_target = os.environ.get('MW_AGENT_SERVICE', '127.0.0.1')


def record_error(error):
    span = trace.get_current_span()
    span.record_exception(error)
    span.set_status(trace.Status(trace.StatusCode.ERROR, str(error)))


# def set_attribute(name, value):
#     if type(name) is not str:
#         print("name must be a string")
#         return
#     if type(value) is not str:
#         print("value must be a string")
#         return
#     span = trace.get_current_span()
#     span.set_attribute(name, value)



import os
import logging
from _meter import collect_metrics
from _tracer import record_error
from _logger import log_handler
from _profiler import collect_profiling

mw_agent_target = os.environ.get('MW_AGENT_SERVICE', '127.0.0.1')


class MwTracker:
    def __init__(self, access_token=None):
        pid = os.getpid()
        os.environ["OTEL_TRACES_EXPORTER"] = "otlp"
        os.environ["OTEL_METRICS_EXPORTER"] = "otlp"
        os.environ["OTEL_LOGS_EXPORTER"] = "otlp"

        # Way 1:
        # project_name = project_name or f"Project-{pid}"
        # service_name = service_name or f"Service-{pid}"
        # access_token = access_token or ""

        # os.environ["OTEL_SERVICE_NAME"] = service_name
        # os.environ["OTEL_RESOURCE_ATTRIBUTES"] = f"project.name={project_name},mw.app.lang=python,runtime.metrics.python=true"

        # Way 2:
        resource_attributes = os.environ.get("OTEL_RESOURCE_ATTRIBUTES")
        _service_name = os.environ.get("OTEL_SERVICE_NAME")
        _project_name = self._get_project_name(resource_attributes)

        project_name = _project_name or f"Project-{pid}"
        service_name = _service_name or f"Service-{pid}"
        access_token = access_token or ""

        # Set values in self
        self.project_name = project_name
        self.service_name = service_name
        self.access_token = access_token

    # For Metrics
    def collect_metrics(self):
        collect_metrics()

    # For Profiling
    def collect_profiling(self) -> None:
        collect_profiling(self.service_name, self.access_token)

    # For Logging
    def collect_logs(self):
        handler = log_handler(self.project_name, self.service_name)
        logging.getLogger().addHandler(handler)
        logging.setLogRecordFactory(self._set_custom_log_attr)

    # For Tracing
    def record_error(self, error):
        record_error(error)

    # For Utils
    def _set_custom_log_attr(self, *args, **kwargs):
        record = logging.LogRecord(*args, **kwargs)
        record.__dict__.update({
            "project.name": self.project_name,
            "service.name": self.service_name,
            "mw.app.lang": "python"
        })
        return record

    def _get_project_name(self, resource_attributes) -> str:
        project_name = None
        if resource_attributes is not None:
            attributes_list = resource_attributes.split(',')
            for attribute in attributes_list:
                key, value = attribute.split('=')
                if key == 'project.name':
                    project_name = value

        return project_name



import os
import requests, json
import pyroscope

mw_agent_target = os.environ.get('MW_AGENT_SERVICE', '127.0.0.1')


def collect_profiling(service_name, access_token="") -> None:
    if access_token != "":

        # Setting Middleware Account Authentication URL
        auth_url = os.getenv('MW_AUTH_URL', 'https://app.middleware.io/api/v1/auth')

        # Setting Middleware Profiling Server URL
        profiling_server_url = os.getenv('MW_PROFILING_SERVER_URL', 'https://profiling.middleware.io')

        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
            # "Authorization": "Bearer " + c.accessToken
            "Authorization": "Bearer " + access_token
        }
        try:
            response = requests.post(auth_url, headers=headers)

            # Checking if auth API returns status code 200
            if response.status_code == 200:
                data = json.loads(response.text)

                # Checking if a tenantID could be fetched from API Key
                if data["success"]:
                    account = data["data"]["account"]
                    pyroscope.configure(
                        application_name=service_name,  # replace this with some name for your application
                        server_address=profiling_server_url,
                        # replace this with the address of your pyroscope server
                        tenant_id=account,
                    )
                else:
                    print("Request failed: " + data["error"])
            else:
                print("Request failed with status code: " + str(response.status_code))
        except Exception as e:
            print("Error making request:", e)
    else:
        print("Profiling is not enabled or access token is empty")



import setuptools

try:
    with open("requirements.txt", "r") as f:
        requirements = f.read().splitlines()
except IOError:
    requirements = []


setuptools.setup(

    name="middleware-apm",
    version="0.2.0rc3",
    install_requires=requirements,
    author="middleware-dev",
    description="This package is use to check the RAM and CPU Usage of Current Device.",
    long_description=open('README.md').read(),
    long_description_content_type='text/plain',
    packages=["mw_tracker"],
    url = "https://github.com/middleware-labs/agent-apm-python.git",
    entry_points={
        'console_scripts': [
            'middleware-instrument = opentelemetry.instrumentation.auto_instrumentation:run',
            'middleware-bootstrap = opentelemetry.instrumentation.bootstrap:run',
        ],
    },
)